#include "cublas_v2.h"

// Required for CUDA check
#include "tensorflow/core/util/port.h"

// GPU allocator
#include "tensorflow/core/common_runtime/gpu/gpu_id.h"
#include "tensorflow/core/common_runtime/gpu/gpu_id_utils.h"
#include "tensorflow/core/common_runtime/gpu/gpu_init.h"
#include "tensorflow/core/common_runtime/gpu/gpu_bfc_allocator.h"

#include "tensorflow/core/common_runtime/direct_session.h"

#include "tensorflow/cc/framework/scope.h"
#include "tensorflow/cc/ops/standard_ops.h"
#include "tensorflow/cc/client/client_session.h"

// CUDA includes. Order matters
#include "dynlink_nvcuvid.h"
#include "cuda_runtime_api.h"

// CUDA kernel function (my)
#include "array.h"

using namespace std;

using tensorflow::CallableOptions;
using tensorflow::DeviceAttributes;
using tensorflow::Scope;
using tensorflow::Session;
using tensorflow::Status;
using tensorflow::Tensor;


Status loadGraph(unique_ptr<tensorflow::Session> *session){

    tensorflow::GraphDef graph_def;

    using namespace tensorflow;
    using namespace tensorflow::ops;

    auto scope = Scope::NewRootScope();

    // TF likes power of 2
    tensorflow::TensorShape shape = tensorflow::TensorShape({256});

    auto a  = Placeholder(scope.WithOpName("array_tensor_in"), DT_UINT8, Placeholder::Shape(shape));

    //auto c0 = Const(scope,{256});
    //auto c1 = Const(scope, (uint8_t)1, {});
    //auto c2  = Fill(scope, c0, c1);

    auto b  = Identity(scope.WithOpName("array_tensor_out"), a);
    //auto b  = Identity(scope.WithOpName("array_tensor_out"), c2);
    //auto b = Add(scope.WithOpName("array_tensor_out"),c2,a);

    TF_CHECK_OK(scope.ToGraphDef(&graph_def));

    //tensorflow::WriteTextProto(Env::Default(), "mygraph.pbtxt", graph_def);

    tensorflow::SessionOptions session_options;
    session_options.config.mutable_gpu_options()->set_allow_growth(true);
    //session_options.config.mutable_gpu_options()->set_per_process_gpu_memory_fraction(0.1);

    session->reset(tensorflow::NewSession(session_options));
    Status session_create_status = (*session)->Create(graph_def);
    if (!session_create_status.ok()){
        LOG(ERROR) << "loadGraph(): ERROR" << session_create_status;
    }
    return Status::OK();
}


bool IsCUDATensor(const Tensor &t){
    cudaPointerAttributes attributes;
    cudaError_t err =
        cudaPointerGetAttributes(&attributes, t.tensor_data().data());
    if (err == cudaErrorInvalidValue)
        return false;
    CHECK_EQ(cudaSuccess, err) << cudaGetErrorString(err);
#if CUDART_VERSION >= 10000
    return (attributes.type == cudaMemoryTypeDevice);
#else
    return (attributes.memoryType == cudaMemoryTypeDevice);
#endif
}


string GPUDeviceName(Session* session) {
  std::vector<DeviceAttributes> devices;
  TF_CHECK_OK(session->ListDevices(&devices));
  for (const DeviceAttributes& d : devices) {
    LOG(INFO) << "Device: " << d.name();
    if (d.device_type() == "GPU" || d.device_type() == "gpu") {
      return d.name();
    }
  }
  return "";
}


int main(int, char**) {

    // check: TF built with CUDA support
    if (!tensorflow::IsGoogleCudaEnabled()){
        LOG(INFO) << "Tensorflow built without CUDA. Rebuild with -c opt --config=cuda";
        return -1;
    }else{
        LOG(INFO) << "Tensorflow built with CUDA, keep running" << endl;
    }

    // check and init CUDA drivers and libs
    void *hHandleDriver = nullptr;
    CUresult cuda_res = cuInit(0, __CUDA_API_VERSION, hHandleDriver);
    if (cuda_res != CUDA_SUCCESS)
    {
        throw exception();
    }
    cuda_res = cuvidInit(0);
    if (cuda_res != CUDA_SUCCESS)
    {
        throw exception();
    }
    LOG(INFO) << "\033[1;32m" << "CUDA init: ok" << "\033[0m";

    std::unique_ptr<tensorflow::Session> session;

    loadGraph(&session);

    const string inputLayer  = "array_tensor_in:0";
    const string outputLayer = "array_tensor_out:0";

    // do the opts
    CallableOptions opts;
    Session::CallableHandle feed_gpu_fetch_cpu;

    opts.add_feed(inputLayer);
    opts.add_fetch(outputLayer);

    const string gpu_device_name = GPUDeviceName(session.get());

    opts.clear_fetch_devices();
    opts.mutable_feed_devices()->insert({inputLayer, gpu_device_name});

    auto runStatus = session->MakeCallable(opts, &feed_gpu_fetch_cpu);
    if (!runStatus.ok())
    {
        LOG(ERROR) << "\033[1;31m" << "Failed to make callable" << "\033[0m";
    }

    // TF likes power of 2 and 256s
    tensorflow::TensorShape shape = tensorflow::TensorShape({256});
    // allocate tensor on the GPU
    tensorflow::PlatformGpuId platform_gpu_id(0);

    tensorflow::GPUMemAllocator *sub_allocator =
        new tensorflow::GPUMemAllocator(
            tensorflow::GpuIdUtil::ExecutorForPlatformGpuId(platform_gpu_id).ValueOrDie(),
            platform_gpu_id, false, {}, {});

    tensorflow::GPUBFCAllocator *allocator =
        new tensorflow::GPUBFCAllocator(sub_allocator, shape.num_elements() * sizeof(tensorflow::DT_UINT8), "GPU_0_bfc");

    auto inputTensor = Tensor(allocator, tensorflow::DT_UINT8, shape);

    LOG(INFO) << "\033[1;37m" << "Is CUDA Tensor? " << (IsCUDATensor(inputTensor)?"\033[1;32myes":"\033[1;31mno") << "\033[0m";

    tensorflow::uint8 *p = inputTensor.flat<tensorflow::uint8>().data();

    // CUDA kernel call
    myCreateCUDAArray(p);

    vector<Tensor> outputs;
    LOG(INFO) << "RunCallable()...";

    runStatus = session->RunCallable(feed_gpu_fetch_cpu, {inputTensor}, &outputs, nullptr);
    if (!runStatus.ok())
    {
        LOG(ERROR) << "Running model failed: " << runStatus;
        return -1;
    }

    LOG(INFO) << "RunCallable() output:";

    LOG(INFO) << outputs[0].DebugString();
    auto tmap = outputs[0].tensor<uint8_t, 1>();

    cout << "\033[1;37m";
    for (int d = 0; d < 256; d++) {
        cout << (int) tmap(d);
        if (d!=255) cout << ", ";
    }
    cout << "\033[0m" << endl;

    session->ReleaseCallable(feed_gpu_fetch_cpu);

    /*
    using namespace tensorflow;
    using namespace tensorflow::ops;

    auto root = Scope::NewRootScope();

    //auto A = Const(root, {1,16,16,1},{8,8});
    //auto B = Const(root, 5,{});

    auto A = Placeholder(root.WithOpName("A"), DT_INT32, Placeholder::Shape({2,2}));

    auto v = Identity(root.WithOpName("v"), A);
    //auto v = Fill(root.WithOpName("v"), A, B);
    // Vector b = [3 5]
    //auto b = Const(root, {{3, 5}});
    // v = Ab^T
    //auto v = MatMul(root.WithOpName("v"), A, b, MatMul::TransposeB(true));

    ClientSession session2(root);
    // Run and fetch v
    int aa[2][2] = {{1,2},{3,4}};
    TF_CHECK_OK(session2.Run({ {A,{{1,2},{3,4}}} },{v}, &outputs));
    // Expect outputs[0] == [19; -3]
    LOG(INFO) << outputs[0].DebugString();
    cout << outputs[0].matrix<int>() << endl;
    */

    return 0;
}
